import java.util.Iterator;
import java.util.NoSuchElementException;

public class SinglyLinkedList<E> implements Iterable<E> {
    private ListNode head;
    private int count;

    private class ListNode {
        private E data;
        private ListNode next;

        ListNode(E data, ListNode next) {
            this.data = data;
            this.next = next;
        }
    }

    public SinglyLinkedList() {
        head = null;
        count = 0;
    }

    public boolean isEmpty() {
        return head == null;
    }

    public int size() {
        return count;
    }

    public void addFirst(E data) {
        // construct a new node to be the head of the list
        // this constructor will set the next field of the new node
        // to refer to the current head
        ListNode newHead = new ListNode(data, head);
        // update the list
        head = newHead;
        count++;
    }

    public E removeFirst() {
        if (isEmpty()) {
            // if the user attempts to remove from an empty list, throw an exception
            // with an appropriate error message
            throw new NoSuchElementException("Cannot remove from an empty list");
        }
        // save the head's data in a temporary variable, so we can return it later
        E data = head.data;
        // update the list
        head = head.next;
        count--;
        // return the data from the removed node
        return data;
    }

    public void addLast(E data) {
        // if the list is empty, addLast is equivalent to addFirst
        // since the loop after this if-statement assumes the list is not empty,
        // we'll use addFirst which can handle an empty list
        if (isEmpty()) {
            addFirst(data);
            return;
        }
        // construct the new node
        ListNode newTail = new ListNode(data, null);
        // find the current tail node
        ListNode current = head;
        while (current.next != null) {
            current = current.next;
        }
        // update the list
        current.next = newTail;
        count++;
    }

    public E removeLast() {
        // if the list is empty or has only one element, removeLast is equivalent to
        // removeFirst
        if (size() <= 1) {
            return removeFirst();
        }

        // find the node right before the current tail node
        ListNode current = head;
        while (current.next.next != null) {
            current = current.next;
        }
        // update the list
        E val = current.data;
        current.next = null;
        count--;
        return val;
    }

    private class SinglyLinkedListIterator implements Iterator<E> {
        ListNode current;

        public SinglyLinkedListIterator() {
            current = head;
        }

        public boolean hasNext() {
            return current != null;
        }

        public E next() {
            E val = current.data;
            current = current.next;
            return val;
        }
    }

    @Override
    public Iterator<E> iterator() {
        return new SinglyLinkedListIterator();
    }
}

